#pragma once

#include <sys/endian.h>
#include <rte_byteorder.h>
#include <rte_ether.h>
#include <rte_flow.h>
#include <rte_ip.h>
#include <rte_mbuf.h>
#include <rte_mbuf_core.h>
#include <rte_net.h>
#include <rte_udp.h>
#include <unistd.h>

#include "defs.hh"

#include <random>

#define IP_DEFTTL 64 /* from RFC 1340. */
#define IP_VERSION 0x40
#define IP_HDRLEN 0x05 /* default IP header length == five 32-bits words. */
#define IP_VHL_DEF (IP_VERSION | IP_HDRLEN)
#define IP_ADDR_FMT_SIZE 15

constexpr static uint32_t MAX_JUMBO_MTU = 9000;
constexpr static uint32_t MAX_STANDARD_MTU = 1500;

static inline int
mtu_to_pkt_size(int mtu)
{
	return mtu + RTE_ETHER_HDR_LEN + RTE_ETHER_CRC_LEN;
}

static inline void 
tx_burst_all(int portid, int txqid, struct rte_mbuf ** tx_bufs, int sz)
{
	int remaining = sz;
	while(remaining > 0) {
		remaining -= rte_eth_tx_burst(
			    portid, txqid, &tx_bufs[sz - remaining],
			    remaining);
	}
}

constexpr static uint32_t ETHER_FRAME_MAGIC = 0xDCDCE5E5;
const static struct rte_ether_addr POU_MAC {
	0x01, 0x00, 0x5e, 0x00, 0x01, 0x81
};
const static uint32_t POU_IP = RTE_IPV4(224, 0, 1, 129);
const static uint16_t POU_PORT = 320;
/* Khat Protocol:
 *   khat only processes two kinds of packets - LOAD and PROBE
 *   rat:
 *        rat -> LOAD -> khat
 *        khat -> LOAD_RESP -> rat
 *   cat:
 *        cat -> PROBE -> khat (cat tx timestamps)
 *        khat -> PROBE_RESP -> cat (cat rx timestamps and khat tx/rx
 * timestamps) khat -> STAT -> cat (khat sends its tx/rx timestamps)
 */

/* Rat Protocol:
 *   cat & rat:
 *      1. both launch with full parameters
 *         rat with slave flag
 *         cat with master flag
 *      2. rats create threads and wait for cat's signal
 *      3. cat creates threads
 *      4. cat -> rats SYNC
 *      5. rats -> cat SYNC_ACK and start running
 *      6. cat start running after received all SYNC_ACKs
 *      7. cat stops running, cat -> rats FIN
 *      8. rats stops running, rats -> cat FIN_ACK with QPS
 *      9. cat exits after receiving all FIN_ACKs and flushing statsGG
 */

struct ptp_hdr {
	uint8_t ptp_msg_type;
	uint8_t ptp_ver;
	uint8_t unused[34];
} __attribute__((packed));

struct pkt_hdr {
	struct rte_ether_hdr eth_hdr;
	struct rte_ipv4_hdr ipv4_hdr;
	struct rte_udp_hdr udp_hdr;
	struct ptp_hdr ptp_hdr;
	uint16_t type;
	uint32_t magic;
	char payload[0];
} __attribute__((packed));

struct net_spec {
	uint32_t ip;
	rte_ether_addr mac_addr;
};

static inline void
pkt_hdr_to_netspec(struct pkt_hdr *pkt, struct net_spec *src,
    uint16_t *src_port, struct net_spec *dst, uint16_t *dst_port)
{
	if (src != nullptr) {
		rte_ether_addr_copy(&pkt->eth_hdr.src_addr, &src->mac_addr);
		src->ip = rte_be_to_cpu_32(pkt->ipv4_hdr.src_addr);
	}

	if (src_port != nullptr) {
		*src_port = rte_be_to_cpu_16(pkt->udp_hdr.src_port);
	}

	if (dst != nullptr) {
		rte_ether_addr_copy(&pkt->eth_hdr.dst_addr, &dst->mac_addr);
		dst->ip = rte_be_to_cpu_32(pkt->ipv4_hdr.dst_addr);
	}

	if (dst_port != nullptr) {
		*dst_port = rte_be_to_cpu_16(pkt->udp_hdr.dst_port);
	}
};

struct conn_spec {
	struct net_spec *src;
	uint16_t src_port;
	struct net_spec *dst;
	uint16_t dst_port;
};

// returns 0 on success
static inline int
str_to_netspec(char *str, struct net_spec *out)
{
	const char *tok = "@";
	char *token;
	char *ptr;
	uint32_t a, b, c, d;

	token = strtok_r(str, tok, &ptr);

	if (token == nullptr ||
	    sscanf(token, "%d.%d.%d.%d", &a, &b, &c, &d) != 4) {
		return -1;
	}

	out->ip = RTE_IPV4(a, b, c, d);

	// mac next
	token = strtok_r(nullptr, tok, &ptr);
	if (token == nullptr ||
	    rte_ether_unformat_addr(token, &out->mac_addr) != 0) {
		return -1;
	}

	return 0;
}

constexpr static uint16_t PKT_TYPE_LOAD = 0;
constexpr static uint32_t LOAD_TYPE_CPU = 0; // arg0 = cpu time in us. arg1 = unused
constexpr static uint32_t LOAD_TYPE_MEM = 1; // arg0 = which thread to access. arg1 = how many cachelines to access
constexpr static uint32_t LOAD_TYPE_MAX = LOAD_TYPE_MEM + 1;
struct pkt_payload_load {
	uint32_t epoch;
	uint32_t type; // type of load
	uint32_t arg0;
	uint32_t arg1;
};

constexpr static uint16_t PKT_TYPE_PROBE = 1;
constexpr static uint16_t PKT_TYPE_LOAD_RESP = 2;
constexpr static uint16_t PKT_TYPE_PROBE_RESP = 3;
struct pkt_payload_epoch {
	uint32_t epoch;
};

constexpr static uint16_t PKT_TYPE_STAT = 4;
struct pkt_payload_stat {
	uint32_t epoch;
	uint64_t hw_rx;
	uint64_t hw_tx;
	uint64_t sw_rx;
	uint64_t sw_tx;
};

constexpr static uint16_t PKT_TYPE_SYNC = 5;
constexpr static uint16_t PKT_TYPE_SYNC_ACK = 6;
constexpr static uint16_t PKT_TYPE_FIN = 7;
constexpr static uint16_t PKT_TYPE_FIN_ACK = 8;
struct pkt_payload_qps {
	uint32_t qps;
	uint32_t recved_pkts;
	uint32_t lost_pkts;
};

constexpr static uint16_t NUM_PKT_TYPES = PKT_TYPE_FIN_ACK + 1;
// for fast packet verification
static const uint32_t expected_payload_size[NUM_PKT_TYPES] {
	sizeof(struct pkt_payload_load),  // LOAD
	sizeof(struct pkt_payload_epoch), // PROBE
	sizeof(struct pkt_payload_epoch), // LOAD_RESP
	sizeof(struct pkt_payload_epoch), // PROBE_RESP
	sizeof(struct pkt_payload_stat),  // STAT
	0,				  // SYNC
	0,				  // SYNC_ACK
	0,				  // FIN
	sizeof(struct pkt_payload_qps)	  // FIN_ACK
};

class rdport_generator {
    private:
	DISALLOW_EVIL_CONSTRUCTORS(rdport_generator);
	constexpr static uint32_t MAX_PORT = 65535;
	uint32_t min_port;
	uint32_t cur;
	std::random_device rd;
	std::default_random_engine gen;
	std::uniform_int_distribution<uint32_t> dist;

    public:
	rdport_generator(uint32_t mport)
	    : min_port(mport)
	    , cur(0)
	    , dist(0, MAX_PORT - min_port)
	{
		gen.seed(get_uptime());
		cur = dist(gen);
	}
	uint16_t next()
	{
		uint16_t ret = ((cur) % (MAX_PORT - min_port)) + min_port;
		cur++;
		return ret;
	}
};

#define NTR_PKT(dep, level, pkt, prefix_fmt, ...)                                                                             \
	ntr(dep, level,                                                                                                       \
	    prefix_fmt                                                                                                        \
	    "src: %d.%d.%d.%d:%d@%02x:%02x:%02x:%02x:%02x:%02x dst: %d.%d.%d.%d:%d@%02x:%02x:%02x:%02x:%02x:%02x type: %d\n", \
	    ##__VA_ARGS__,                                                                                                    \
	    (rte_be_to_cpu_32(pkt->ipv4_hdr.src_addr) >> 24) & 0xff,                                                          \
	    (rte_be_to_cpu_32(pkt->ipv4_hdr.src_addr) >> 16) & 0xff,                                                          \
	    (rte_be_to_cpu_32(pkt->ipv4_hdr.src_addr) >> 8) & 0xff,                                                           \
	    (rte_be_to_cpu_32(pkt->ipv4_hdr.src_addr) >> 0) & 0xff,                                                           \
	    rte_be_to_cpu_16(pkt->udp_hdr.src_port),                                                                          \
	    pkt->eth_hdr.src_addr.addr_bytes[0],                                                                                \
	    pkt->eth_hdr.src_addr.addr_bytes[1],                                                                                \
	    pkt->eth_hdr.src_addr.addr_bytes[2],                                                                                \
	    pkt->eth_hdr.src_addr.addr_bytes[3],                                                                                \
	    pkt->eth_hdr.src_addr.addr_bytes[4],                                                                                \
	    pkt->eth_hdr.src_addr.addr_bytes[5],                                                                                \
	    (rte_be_to_cpu_32(pkt->ipv4_hdr.dst_addr) >> 24) & 0xff,                                                          \
	    (rte_be_to_cpu_32(pkt->ipv4_hdr.dst_addr) >> 16) & 0xff,                                                          \
	    (rte_be_to_cpu_32(pkt->ipv4_hdr.dst_addr) >> 8) & 0xff,                                                           \
	    (rte_be_to_cpu_32(pkt->ipv4_hdr.dst_addr) >> 0) & 0xff,                                                           \
	    rte_be_to_cpu_16(pkt->udp_hdr.dst_port),                                                                          \
	    pkt->eth_hdr.dst_addr.addr_bytes[0],                                                                                \
	    pkt->eth_hdr.dst_addr.addr_bytes[1],                                                                                \
	    pkt->eth_hdr.dst_addr.addr_bytes[2],                                                                                \
	    pkt->eth_hdr.dst_addr.addr_bytes[3],                                                                                \
	    pkt->eth_hdr.dst_addr.addr_bytes[4],                                                                                \
	    pkt->eth_hdr.dst_addr.addr_bytes[5], rte_be_to_cpu_16(pkt->type))

static inline void
print_mac(struct rte_ether_addr *mac)
{
	printf("%x:%x:%x:%x:%x:%x", mac->addr_bytes[0], mac->addr_bytes[1],
	    mac->addr_bytes[2], mac->addr_bytes[3], mac->addr_bytes[4],
	    mac->addr_bytes[5]);
}

static inline void
print_ipv4(uint32_t ip)
{
	printf("%d.%d.%d.%d", (ip >> 24) & 0xff, (ip >> 16) & 0xff,
	    (ip >> 8) & 0xff, (ip >> 0) & 0xff);
}

static inline void
dump_pkt(struct rte_mbuf *pkt)
{
	if (rte_pktmbuf_data_len(pkt) < sizeof(struct rte_ether_hdr)) {
		return;
	}

	struct rte_ether_hdr _eth_hdr;
	auto eth_hdr = (struct rte_ether_hdr *)rte_pktmbuf_read(
	    pkt, 0, sizeof(struct rte_ether_hdr), &_eth_hdr);
	if (eth_hdr == nullptr) {
		return;
	}

	// ethernet frame
	printf(
	    "Packet %p: Length 0x%x\n", (void *)pkt, rte_pktmbuf_data_len(pkt));
	printf("    Ethernet header:\n");
	printf("        Src:");
	print_mac(&eth_hdr->src_addr);
	printf("\n");
	printf("        Dst:");
	print_mac(&eth_hdr->dst_addr);
	printf("\n");
	printf("        Type: 0x%x\n", rte_be_to_cpu_16(eth_hdr->ether_type));

	uint16_t ether_type = rte_be_to_cpu_16(eth_hdr->ether_type);
	if (ether_type != RTE_ETHER_TYPE_IPV4) {
		return;
	}

	if (rte_pktmbuf_data_len(pkt) <
	    sizeof(struct rte_ether_hdr) + sizeof(struct rte_ipv4_hdr)) {
		return;
	}

	// dump ip header
	auto ipv4_hdr = (struct rte_ipv4_hdr *)(eth_hdr + 1);
	printf("    IPv4 header:\n");
	printf("        Src:");
	print_ipv4(rte_be_to_cpu_32(ipv4_hdr->src_addr));
	printf("\n");
	printf("        Dst:");
	print_ipv4(rte_be_to_cpu_32(ipv4_hdr->dst_addr));
	printf("\n");
	printf("        Protocol: 0x%x\n", ipv4_hdr->next_proto_id);
}

static inline bool
is_l2ts_pkt(uint16_t type)
{
	return type == PKT_TYPE_PROBE || type == PKT_TYPE_PROBE_RESP;
}

// fills the packet with the information except for the payload itself
static inline struct pkt_hdr *
construct_pkt_hdr(
    struct rte_mbuf *buf, uint16_t type, const struct conn_spec *conn, int pkt_pad_sz)
{
	rte_pktmbuf_reset(buf);

	int total_sz = sizeof(struct pkt_hdr) +
	    expected_payload_size[type];

	if (pkt_pad_sz > total_sz) {
		total_sz = pkt_pad_sz;
	}

	auto pkt_data = (struct pkt_hdr *)rte_pktmbuf_append(buf, total_sz);
	if (pkt_data == nullptr)
		return nullptr;

	struct rte_ether_hdr *eth_hdr;
	struct rte_ipv4_hdr *ipv4_hdr;
	struct rte_udp_hdr *udp_hdr;
	bool is_ts_pkt = is_l2ts_pkt(type);

	// single segment
	buf->nb_segs = 1;

	// construct l2 header
	eth_hdr = &pkt_data->eth_hdr;
	rte_ether_addr_copy(&conn->src->mac_addr, &eth_hdr->src_addr);
	if (is_ts_pkt) {
		rte_ether_addr_copy(&POU_MAC, &eth_hdr->dst_addr);
	} else {
		rte_ether_addr_copy(&conn->dst->mac_addr, &eth_hdr->dst_addr);
	}
	eth_hdr->ether_type = rte_cpu_to_be_16(RTE_ETHER_TYPE_IPV4);
	buf->l2_len = sizeof(struct rte_ether_hdr);

	// construct l3 header
	ipv4_hdr = &pkt_data->ipv4_hdr;
	memset(ipv4_hdr, 0, sizeof(struct rte_ipv4_hdr));
	ipv4_hdr->version_ihl = IP_VHL_DEF;
	ipv4_hdr->type_of_service = 0;
	ipv4_hdr->fragment_offset = 0;
	ipv4_hdr->time_to_live = IP_DEFTTL;
	ipv4_hdr->next_proto_id = IPPROTO_UDP;
	ipv4_hdr->packet_id = 0;
	ipv4_hdr->src_addr = rte_cpu_to_be_32(conn->src->ip);
	if (is_ts_pkt) {
		ipv4_hdr->dst_addr = rte_cpu_to_be_32(POU_IP);
	} else {
		ipv4_hdr->dst_addr = rte_cpu_to_be_32(conn->dst->ip);
	}
	ipv4_hdr->total_length = rte_cpu_to_be_16(total_sz - sizeof(struct rte_ether_hdr) - sizeof(struct rte_ipv4_hdr));
	ipv4_hdr->hdr_checksum = 0;
	buf->l3_len = sizeof(struct rte_ipv4_hdr);

	// construct l4 header
	udp_hdr = &pkt_data->udp_hdr;
	udp_hdr->src_port = rte_cpu_to_be_16(conn->src_port);
	if (is_ts_pkt) {
		udp_hdr->dst_port = rte_cpu_to_be_16(POU_PORT);
	} else {
		udp_hdr->dst_port = rte_cpu_to_be_16(conn->dst_port);
	}
	udp_hdr->dgram_cksum = 0; /* No UDP checksum. */
	udp_hdr->dgram_len = total_sz - sizeof(struct rte_ether_hdr) - sizeof(struct rte_ipv4_hdr) - sizeof(struct rte_udp_hdr);
	buf->l4_len = sizeof(struct rte_udp_hdr);
	buf->ol_flags |= RTE_MBUF_F_TX_IPV4;
	buf->ol_flags |= RTE_MBUF_F_TX_IP_CKSUM;
	buf->ol_flags |= RTE_MBUF_F_TX_UDP_CKSUM;

	if (is_ts_pkt) {
		// set misc flags
		buf->ol_flags |= RTE_MBUF_F_TX_IEEE1588_TMST;
		pkt_data->ptp_hdr.ptp_ver = 0x2;      // VER 2
		pkt_data->ptp_hdr.ptp_msg_type = 0x0; // SYNC
	} else {
		pkt_data->ptp_hdr.ptp_ver = 0xff; // invalid ver
	}

	pkt_data->type = rte_cpu_to_be_16(type);
	pkt_data->magic = rte_cpu_to_be_32(ETHER_FRAME_MAGIC);

	return pkt_data;
}

// returns 0 on success
static inline int
alloc_pkt_hdr(struct rte_mempool *pool, uint16_t type,
    const struct conn_spec *conn, int pkt_pad_sz, struct rte_mbuf **mbuf_out,
    struct pkt_hdr **hdr_out)
{
	struct pkt_hdr *hdr;
	struct rte_mbuf *pkt = rte_pktmbuf_alloc(pool);
	if (pkt == nullptr) {
		return -1;
	}

	// printf("alloc_pkt_hdr:\n");
	// printf("from ");
	// print_mac(&conn->src->mac_addr);
	// printf("\nto ");
	// print_mac(&conn->dst->mac_addr);
	// printf("\n");

	hdr = construct_pkt_hdr(pkt, type, conn, pkt_pad_sz);
	if (hdr == nullptr) {
		rte_pktmbuf_free(pkt);
		return -1;
	}

	*mbuf_out = pkt;
	*hdr_out = hdr;
	return 0;
}

static inline struct pkt_hdr *
check_valid_packet(struct rte_mbuf *pkt, const struct rte_ether_addr *host_mac)
{
	struct pkt_hdr *pkt_data = nullptr;
	const struct rte_ether_addr *expected_mac = nullptr;
	uint16_t type;
	const uint32_t data_len = rte_pktmbuf_data_len(pkt);

	if (data_len < sizeof(struct pkt_hdr)) {
		return nullptr;
	}

	pkt_data = rte_pktmbuf_mtod(pkt, struct pkt_hdr *);

	// check MAGIC
	if (rte_be_to_cpu_32(pkt_data->magic) != ETHER_FRAME_MAGIC) {
		return nullptr;
	}

	type = rte_be_to_cpu_16(pkt_data->type);
	// check type and payload size
	if ((type >= NUM_PKT_TYPES) ||
	    (data_len <
		(sizeof(struct pkt_hdr) +
		    expected_payload_size[rte_be_to_cpu_16(pkt_data->type)]))) {
		return nullptr;
	}

	// strict dest mac filter
	if (host_mac != nullptr) {
		if (is_l2ts_pkt(type)) {
			// dst mac must be the broadcast addr
			expected_mac = &POU_MAC;
		} else {
			// dst mac must match the host mac
			expected_mac = host_mac;
		}

		if (!rte_is_same_ether_addr(
			expected_mac, &pkt_data->eth_hdr.dst_addr))
			return nullptr;
	}

	return pkt_data;
}
