#include "net/netsup.hh"
#include <cstdlib>

#include "rte_build_config.h"
#include "rte_common.h"
#include "rte_config.h"
#include "rte_ether.h"
#include "rte_lcore.h"
#include "rte_mempool.h"
#include "rte_mbuf.h"
#include "rte_errno.h"
#include "rte_ethdev.h"

#include "ntr.h"

static struct rte_mempool *g_mempools[MAX_NUMA_NODES] = {nullptr};
static unsigned int g_mempool_sz = 0;

static void
mempool_init(struct mem_conf *mconf)
{
	struct rte_mempool * mbuf_pool;
	char mempool_name[64];

	for (int i = 0; i < (int)rte_socket_count(); i++) {
		uint32_t nodeid = i;
		// ntr(NTR_DEP_USER1, NTR_LEVEL_INFO,
		//     "mempool_init: creating mempool for node %d\n", nodeid);

		// create one mbuf pool per socket
		snprintf(mempool_name, sizeof(mempool_name), "net_mempool_%d", nodeid);

		mbuf_pool = rte_pktmbuf_pool_create(mempool_name, mconf->num_elements, 
					mconf->cache_size, mconf->priv_size, 
				mconf->data_room_size, nodeid);

		if (mbuf_pool == nullptr) {
			rte_exit(EXIT_FAILURE, "cannot create mbuf pool: %d\n", rte_errno);
		}

		g_mempools[nodeid] = mbuf_pool;
		g_mempool_sz++;
		ntr(NTR_DEP_USER1, NTR_LEVEL_INFO, "mempool_init: created mempool for node %d\n", nodeid);
	}
}

struct rte_mempool *
mempool_get(int nodeid)
{
	if ((unsigned int)nodeid < g_mempool_sz) {
		return g_mempools[nodeid];
	}
	return nullptr;
}

static void
port_init(struct device_conf *dconf)
{
	struct rte_ether_addr addr;
	struct rte_eth_dev_info dev_info {
	};
	struct rte_eth_conf port_conf;
	struct rte_eth_txconf txconf {
	};
	struct rte_eth_rxconf rxconf {
	};
	int ret;

	int num_threads = CPU_COUNT(&dconf->core_affinity);
	if (rte_eth_dev_count_avail() == 0) {
		rte_exit(EXIT_FAILURE, "number of ports must be > 0\n");
	}

	if (!rte_eth_dev_is_valid_port(dconf->portid)) {
		rte_exit(EXIT_FAILURE, "cannot find port %d\n", dconf->portid);
	}

	if ((ret = rte_eth_macaddr_get(dconf->portid, &addr)) != 0) {
		rte_exit(EXIT_FAILURE, "cannot get mac address of port: %d\n", ret);
	}

	ret = rte_eth_dev_info_get(dconf->portid, &dev_info);
	if (ret != 0) {
		rte_exit(EXIT_FAILURE, "failed to get dev info: %d", ret);
	}

	memset(&port_conf, 0, sizeof(struct rte_eth_conf));
	port_conf.rxmode.mtu = dconf->mtu;
	port_conf.rxmode.mq_mode = RTE_ETH_MQ_RX_RSS;
	port_conf.rx_adv_conf.rss_conf.rss_key = nullptr;
	port_conf.rx_adv_conf.rss_conf.rss_hf = dconf->rss_hf;

	port_conf.rxmode.offloads = dconf->rx_offloads;
	port_conf.txmode.offloads = dconf->tx_offloads;

	/* Configure the Ethernet device. */
	ret = rte_eth_dev_configure(dconf->portid, num_threads, num_threads, &port_conf);
	if (ret != 0)
		rte_exit(EXIT_FAILURE, "failed to configure port: %d\n", ret);

	ret = rte_eth_dev_adjust_nb_rx_tx_desc(dconf->portid, &dconf->rx_ring_sz, &dconf->tx_ring_sz);
	if (ret != 0)
		rte_exit(EXIT_FAILURE, "failed to set rx tx queue size: %d\n", ret);

	/* Allocate and set up 1 RX queue per thread per Ethernet port. */
	rxconf = dev_info.default_rxconf;
	rxconf.offloads = port_conf.rxmode.offloads;
	rxconf.rx_nseg = 0;
	rxconf.rx_seg = nullptr;
	txconf = dev_info.default_txconf;
	txconf.offloads = port_conf.txmode.offloads;

	int core;
	int qid = 0;
	CPU_FOREACH_ISSET(core, &dconf->core_affinity) {
		int socket = rte_lcore_to_socket_id(core);
		ntr(NTR_DEP_USER1, NTR_LEVEL_INFO, "port_init: setting up rx & tx queue for core %d (socket %d)...\n", core, socket);
		ret = rte_eth_rx_queue_setup(dconf->portid, qid, dconf->rx_ring_sz, socket, &rxconf, mempool_get(socket));
		if (ret < 0)
			rte_exit(EXIT_FAILURE, "failed to setup rx queue for core %d: %d\n", core, ret);
		
		ret = rte_eth_tx_queue_setup(dconf->portid, qid, dconf->tx_ring_sz, socket, &txconf);
		if (ret < 0)
			rte_exit(EXIT_FAILURE, "failed to setup tx queue for core %d: %d", core, ret);

		qid++;
	}

	// set mtu
	ret = rte_eth_dev_set_mtu(dconf->portid, dconf->mtu);
	if (ret != 0)
		rte_exit(EXIT_FAILURE, "failed to set mtu: %d\n", ret);

	ret = rte_eth_dev_start(dconf->portid);
	if (ret < 0)
		rte_exit(EXIT_FAILURE, "failed to start port: %d\n", ret);

	if (dconf->timesync) {
		ret = rte_eth_timesync_enable(dconf->portid);
		if (ret != 0)
			rte_exit(EXIT_FAILURE, "failed to enable timesync: %d\n", ret);
	}

	/* Enable RX in promiscuous mode for the Ethernet device. */
	ret = rte_eth_promiscuous_enable(dconf->portid);
	if (ret != 0)
		rte_exit(EXIT_FAILURE, "failed to enable promiscuous mode: %d\n", ret);

	for (int i = 0; i < num_threads; i++) {
		if (dconf->tx_fn != nullptr) {
			if (rte_eth_add_tx_callback(dconf->portid, i, dconf->tx_fn, dconf->tx_user) == nullptr) {
				rte_exit(EXIT_FAILURE, "failed to attach callback to tx queue %d\n", i);
			}
		}
		
		if (dconf->rx_fn != nullptr) {
			if (rte_eth_add_rx_callback(dconf->portid, i, dconf->rx_fn, dconf->rx_user) == nullptr) {
				rte_exit(EXIT_FAILURE, "failed to attach callback to rx queue %d\n", i);
			}
		}
	}

	// sync_port_clock(portid);

	ntr(NTR_DEP_USER1, NTR_LEVEL_INFO, 
	"port_init: configured port %d on socket %d with mac addr %x:%x:%x:%x:%x:%x\n",
		dconf->portid, rte_eth_dev_socket_id(dconf->portid),
		addr.addr_bytes[0],
		addr.addr_bytes[1],
		addr.addr_bytes[2],
		addr.addr_bytes[3],
		addr.addr_bytes[4],
		addr.addr_bytes[5]);
}

void
dpdk_init(struct device_conf *dconf, struct mem_conf *mconf)
{
	if (rte_socket_count() > (int)MAX_NUMA_NODES) {
		rte_exit(EXIT_FAILURE, "too many numa nodes\n");
	}

	// ensure 1-1 mapping
	for (int i = 0; i < (int)rte_socket_count(); i++) {
		if (rte_socket_id_by_idx(i) != i) {
			rte_exit(EXIT_FAILURE, "socket %d has id %d instead.\n", i, rte_socket_id_by_idx(i));
		}
	}

	mempool_init(mconf);

	port_init(dconf);
}

void
dpdk_cleanup(struct device_conf * dconf)
{
	rte_eth_dev_stop(dconf->portid);
	rte_eth_dev_close(dconf->portid);

	for (int i = 0; i < (int)rte_socket_count(); i++) {
		rte_mempool_free(g_mempools[i]);
	}
}