import json
import numpy as np

class iperf_json_parser:
    def __init__(self, inputs):
        self.aggregate_egress_bps = 0
        self.jsonobjs = []
        for input in inputs:
            jsobj = json.loads(input)
            self.jsonobjs.append(jsobj)
            each_bps = jsobj['end']['sum_sent']['bits_per_second']
            self.aggregate_egress_bps += each_bps

class memloadgen_parser:
    def __init__(self, input, min, max):
        lines = input.split('\n')
        if max > len(lines):
            max = len(lines)
        if len(lines) <= min:
            raise Exception("Not enough lines!")
        if min > max:
            min = max
        arr = []
        for i in range(min, max):
            arr.append(int(lines[i]))
        self.bps = np.mean(arr)


class pmc_parser:
    def __init__(self, input):
        self.raw = input
        lines = input.split('\n')
        if len(lines) < 2:
            raise Exception("Invalid pmc file format")
        
        spec = lines[0].strip()
        if (spec[0] != '#'):
            raise Exception("Invalid pmc file spec line: \"" + lines[0] + "\"")
        spec = spec.split(' ')
        self.cores = len(spec) - 1
        elements = spec[1].split('/')
        if (len(elements) != 3):
            raise Exception("Invalid pmc file spec line: \"" + lines[0] + "\"")
        self.counter = elements[2].strip()

        last_line = lines[-1]
        elements = last_line.split(' ')
        total = 0
        for e in elements:
            if (len(e) > 0):
                total += int(e)
        self.count = total

class khat_parser:
    class pt:
        def __init__(self):
            self.s_htx = 0
            self.s_hrx = 0
            self.s_stx = 0
            self.s_srx = 0
            self.c_htx = 0
            self.c_hrx = 0
            self.c_stx = 0
            self.c_srx = 0
            self.master_total = 0
            self.master_loss = 0
            self.slave_total = 0
            self.slave_loss = 0
            self.qps = 0

    def __init__(self):
        self.datapt = []
        self.srv_hwlat = []
        self.srv_swlat = []
        self.clt_hwlat = []
        self.clt_swlat = []
        self.lat_idx_arr = []
        self.lat_idx_arr.append(self.srv_hwlat)
        self.lat_idx_arr.append(self.srv_swlat)
        self.lat_idx_arr.append(self.clt_hwlat)
        self.lat_idx_arr.append(self.clt_swlat)

    def get_stat_arr(self, idx : int):
        return self.lat_idx_arr[idx]

    def parse(self, output : str):
        first = True
        for line in output.splitlines():
            # the first line is qps
            cells = line.split(',')
            if (first):
                if len(cells) != 5:
                    raise Exception("Invalid headline:" + line)
                self.qps = int(cells[0])
                self.master_recv = int(cells[1])
                self.master_loss = int(cells[2])
                self.slave_recv = int(cells[3])
                self.slave_loss = int(cells[4])
                first = False
                continue
            if len(cells) != 8:
                raise Exception("Invalid line:" + line)
            pt = self.pt()
            pt.c_srx = int(cells[0])
            pt.c_stx = int(cells[1])
            pt.c_hrx = int(cells[2])
            pt.c_htx = int(cells[3])
            pt.s_srx = int(cells[4])
            pt.s_stx = int(cells[5])
            pt.s_hrx = int(cells[6])
            pt.s_htx = int(cells[7])
            self.datapt.append(pt)
            self.srv_hwlat.append(pt.s_htx - pt.s_hrx)
            self.srv_swlat.append(pt.s_stx - pt.s_srx)
            self.clt_hwlat.append(pt.c_hrx - pt.c_htx)
            self.clt_swlat.append(pt.c_srx - pt.c_stx)
            

class mutilate_data:
    def __init__(self):
        self.dat = {}
        self.qps = 0

    def to_string(self):
        ret = "Throughput: " + str(self.qps) + "\n" + json.dumps(self.dat)
        return ret

    @staticmethod
    def parse_mut_output(output):
        ret = mutilate_data()
        succ_qps = False
        succ_read = False
        table = [None, "avg", "std", "min", "5th", "10th", "50th", "90th", "95th", "99th"]
        table_legacy = [None, "avg", "std", "min", "5th", "10th", "90th", "95th", "99th"]
        for line in output.splitlines():
            if line.find("Total QPS") != -1:
                spl = line.split()
                if len(spl) == 7:
                    ret.qps = float(spl[3])
                    succ_qps = True
                else:
                    break
            elif line.find("read") != -1:
                spl = line.split()
                if len(spl) == 10:
                    for i in range(1, len(spl)):
                        ret.dat[table[i]] = float(spl[i])
                    succ_read = True
                elif len(spl) == 9:
                    for i in range(1, len(spl)):
                        ret.dat[table_legacy[i]] = float(spl[i])
                    succ_read = True
                else:
                    break
        
        if not (succ_qps and succ_read):
            raise Exception("Failed to parse data")

        return ret

    @staticmethod
    def parse_mut_sample(fn):
        f = open(fn, "r")
        qps = []
        lat = []
        lines = f.readlines()
        for line in lines:
            entry = line.split()
            if len(entry) != 2:
                raise Exception("Unrecognized line: " + line)
            qps.append(float(entry[0]))
            lat.append(float(entry[1]))
        f.close()
        return qps, lat


    # generate mutilate output format
    @staticmethod
    def build_mut_output(lat_arr, qps_arr):
        output = '{0: <10}'.format('#type') + '{0: >10}'.format('avg') + '{0: >10}'.format('std') + \
                        '{0: >10}'.format('min') + '{0: >10}'.format('5th') + '{0: >10}'.format('10th') + \
                        '{0: >10}'.format('50th') + '{0: >10}'.format('90th')  + '{0: >10}'.format('95th') + '{0: >10}'.format('99th') + "\n"
        
        output += '{0: <10}'.format('read') + '{0: >10}'.format("{:.1f}".format(np.mean(lat_arr))) + ' ' + \
                        '{0: >10}'.format("{:.1f}".format(np.std(lat_arr))) + ' ' + \
                        '{0: >10}'.format("{:.1f}".format(np.min(lat_arr))) + ' ' + \
                        '{0: >10}'.format("{:.1f}".format(np.percentile(lat_arr, 5))) + ' ' + \
                        '{0: >10}'.format("{:.1f}".format(np.percentile(lat_arr, 10))) + ' ' + \
                        '{0: >10}'.format("{:.1f}".format(np.percentile(lat_arr, 50))) + ' ' + \
                        '{0: >10}'.format("{:.1f}".format(np.percentile(lat_arr, 90))) + ' ' + \
                        '{0: >10}'.format("{:.1f}".format(np.percentile(lat_arr, 95))) + ' ' + \
                        '{0: >10}'.format("{:.1f}".format(np.percentile(lat_arr, 99))) + ' ' + "\n" \

        output += "\n" + "Total QPS = " + "{:.1f}".format(np.mean(qps_arr)) + " (0 / 0s)"

        return output