#!/usr/bin/env python3.6

import numpy as np
import sys
import re
import os
import json
import getopt
import math
import concurrent.futures as CF

columns = [
    ("Req per second", "rps", ".2f"),
    ("Bytes per second", "bps", ".2f"),
    ("Average Latency", "lat_avg", ".2f"),
    ("50th Latency", "lat_50", ".0f"),
    ("95th Latency", "lat_95", ".0f"),
    ("99th Latency", "lat_99", ".0f"),
    ("Latency stddev", "lat_std", ".2f")
]

TIME = 30
REQ_SZ = 4096

class DatObj:
    def __init__(self, raw : list, time : int, req_sz : int):
        self.raw = raw
        self.rps = len(raw) / time
        self.bps = self.rps * req_sz
        self.lat_avg = np.average(self.raw)
        self.lat_99 = np.percentile(self.raw, 99)
        self.lat_95 = np.percentile(self.raw, 95)
        self.lat_50 = np.percentile(self.raw, 50)
        self.lat_std = np.std(self.raw)

def parse_file(lines : list, time : int, req_sz : int) -> DatObj :
    raw = []
    for line in lines:
        if len(line) > 0:
            raw.append(int(line))
    return DatObj(raw, time, req_sz)

def output_col():
    ret = "Benchmark"
    for name,_,_ in columns:
        ret = ret + "," + name + "," + name + " (NUMA)" + "," + "% change"
    return ret

def get_attr_or_none(obj, attr):
    if (obj != None):
        val = getattr(obj, attr)
    else:
        val = None
    return val

def output_objs(name: str, obj : DatObj, obj_numa : DatObj):
    ret = name
    for _, attr, fmt in columns:
        val = get_attr_or_none(obj, attr)
        val_numa = get_attr_or_none(obj_numa, attr)
 
        ret = ret + "," + (format(val, fmt) if val != None else "N/A") 
        ret = ret + "," + (format(val_numa, fmt) if val_numa != None else "N/A") 
        
        if val == None or val_numa == None:
            ret = ret + "," + "N/A"
        else:
            ret = ret + "," + format((val_numa - val) / val * 100, ".2f") + "%"
    return ret

def process_file(f : str, obj_map):
    with open(f, "r") as fp:
        lines = fp.readlines()
    
    bench_name = os.path.basename(f)
    obj_map[bench_name] = parse_file(lines, TIME, REQ_SZ)
    print("Processed file " + f + ". Benchmark name: " + bench_name)

def process_dir(path : str, obj_map):
    files = [os.path.abspath(os.path.join(path, x)) for x in os.listdir(path)]
    for f in files:
        if (".sh" in f):
            continue
        if (os.path.isfile(f)):
            process_file(f, obj_map)

def main():
    datdir = None
    options = getopt.getopt(sys.argv[1:], 'd:')[0]

    for opt, arg in options:
        if opt in ('-d'):
            datdir = arg

    if datdir == None:
        raise Exception("Must specify -d parameter")

    obj_map = dict()
    process_dir(datdir, obj_map)

    with open("results.csv", "w") as f:
        f.write(output_col())
        f.write("\n")

        for bench in obj_map:
            if bench.endswith("_numa"):
                continue
            f.write(output_objs(bench, obj_map[bench], obj_map.get(bench+"_numa")))
            f.write("\n")

if __name__ == "__main__":
    main()